package io.educative.soap.helpers;

import org.springframework.ws.client.core.WebServiceMessageCallback;
import org.springframework.ws.client.core.WebServiceTemplate;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.dataformat.xml.XmlMapper;

import io.educative.soap.AuthHeaderRequestCallback;
import io.educative.soap.utils.ConfigManager;
import io.educative.soap_automation.CreateStudentRequest;
import io.educative.soap_automation.CreateStudentResponse;
import io.educative.soap_automation.DeleteStudentRequest;
import io.educative.soap_automation.DeleteStudentResponse;
import io.educative.soap_automation.GetStudentsRequest;
import io.educative.soap_automation.GetStudentsResponse;
import io.educative.soap_automation.UpdateStudentRequest;
import io.educative.soap_automation.UpdateStudentResponse;
import io.qameta.allure.Allure;
import io.qameta.allure.Step;

public class StudentServiceHelper {

	private final WebServiceTemplate webServiceTemplate;

	private final boolean useAuth;

	private static final String SERVICE_URL = ConfigManager.getInstance().getString("api.base_url");

	private static final String USERNAME = ConfigManager.getInstance().getString("api.auth.username");

	private static final String PASSWORD = ConfigManager.getInstance().getString("api.auth.password");

	public StudentServiceHelper(WebServiceTemplate webServiceTemplate, boolean useAuth) {
		this.webServiceTemplate = webServiceTemplate;
		this.useAuth = useAuth;
	}

	public StudentServiceHelper(WebServiceTemplate webServiceTemplate) {
		this(webServiceTemplate, false);
	}

	@Step
	public CreateStudentResponse createStudent(String firstName, String lastName, String gender) {

		// Creating Student with following details
		CreateStudentRequest request = new CreateStudentRequest();
		request.setGender(gender);
		request.setFirstName(firstName);
		request.setLastName(lastName);

		addAttachment(request);
		CreateStudentResponse response = (CreateStudentResponse) webServiceTemplate.marshalSendAndReceive(SERVICE_URL,
				request,
				getAuthRequestCallback());
		addAttachment(response);
		return response;
	}

	@Step
	public GetStudentsResponse getStudentById(long id) {

		GetStudentsRequest request = new GetStudentsRequest();
		request.setId(id);

		return getStudents(request);
	}

	@Step
	public GetStudentsResponse getAllStudents() {
		return getStudents(new GetStudentsRequest());
	}

	@Step
	public GetStudentsResponse getStudents(GetStudentsRequest request) {
		addAttachment(request);
		GetStudentsResponse response = (GetStudentsResponse) webServiceTemplate.marshalSendAndReceive(SERVICE_URL,
				request, getAuthRequestCallback());
		addAttachment(response);
		return response;
	}

	private WebServiceMessageCallback getAuthRequestCallback() {
		return useAuth ? null : new AuthHeaderRequestCallback(USERNAME, PASSWORD);
	}

	@Step
	public UpdateStudentResponse updateStudent(long id, String firstName, String lastName, String gender) {

		// Update the Student with information
		UpdateStudentRequest request = new UpdateStudentRequest();
		request.setId(id);
		request.setGender(gender);
		request.setFirstName(firstName);
		request.setLastName(lastName);

		addAttachment(request);
		UpdateStudentResponse response = (UpdateStudentResponse) webServiceTemplate.marshalSendAndReceive(SERVICE_URL,
				request,
				getAuthRequestCallback());
		addAttachment(response);
		return response;
	}

	@Step
	public DeleteStudentResponse deleteStudent(long id) {
		DeleteStudentRequest request = new DeleteStudentRequest();
		request.setId(id);

		addAttachment(request);
		DeleteStudentResponse response = (DeleteStudentResponse) webServiceTemplate.marshalSendAndReceive(SERVICE_URL,
				request,
				getAuthRequestCallback());
		addAttachment(response);
		return response;
	}

	// add attachment to allure report
	private void addAttachment(Object obj) {
		try {
			Allure.getLifecycle().addAttachment(obj.getClass().getSimpleName(), "text/xml", ".xml",
					new XmlMapper().writerWithDefaultPrettyPrinter().writeValueAsBytes(obj));
		} catch (JsonProcessingException e) {
			e.printStackTrace();
		}
	}
}
