package io.educative.soap.server;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import javax.annotation.PostConstruct;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.ExitCodeGenerator;
import org.springframework.boot.SpringApplication;
import org.springframework.boot.WebApplicationType;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.boot.builder.SpringApplicationBuilder;
import org.springframework.boot.web.servlet.ServletRegistrationBean;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ConfigurableApplicationContext;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.io.ClassPathResource;
import org.springframework.stereotype.Component;
import org.springframework.ws.config.annotation.EnableWs;
import org.springframework.ws.config.annotation.WsConfigurerAdapter;
import org.springframework.ws.server.EndpointInterceptor;
import org.springframework.ws.server.endpoint.annotation.Endpoint;
import org.springframework.ws.server.endpoint.annotation.PayloadRoot;
import org.springframework.ws.server.endpoint.annotation.RequestPayload;
import org.springframework.ws.server.endpoint.annotation.ResponsePayload;
import org.springframework.ws.soap.security.xwss.XwsSecurityInterceptor;
import org.springframework.ws.transport.http.MessageDispatcherServlet;
import org.springframework.ws.wsdl.wsdl11.DefaultWsdl11Definition;
import org.springframework.xml.xsd.SimpleXsdSchema;
import org.springframework.xml.xsd.XsdSchema;

import io.educative.soap_automation.CreateStudentRequest;
import io.educative.soap_automation.CreateStudentResponse;
import io.educative.soap_automation.DeleteStudentRequest;
import io.educative.soap_automation.DeleteStudentResponse;
import io.educative.soap_automation.GetStudentsRequest;
import io.educative.soap_automation.GetStudentsResponse;
import io.educative.soap_automation.Student;
import io.educative.soap_automation.UpdateStudentRequest;
import io.educative.soap_automation.UpdateStudentResponse;

@SpringBootApplication
public class SOAPApplication {

	public static void startServer(String... args) {
		SpringApplication.run(SOAPApplication.class, args);
	}

	public static void stopServer() {
		ConfigurableApplicationContext ctx = new SpringApplicationBuilder(SOAPApplication.class)
				.web(WebApplicationType.NONE).run();
		SpringApplication.exit(ctx, new ExitCodeGenerator() {
			@Override
			public int getExitCode() {
				return 0;
			}
		});
	}

	public static void main(String[] args) {
		startServer();
	}
}

@EnableWs
@Configuration
class WebServiceConfig extends WsConfigurerAdapter {

	private static final String SCHEMA_XSD = "students.xsd";
	private static final String TEST_USERNAME = "testuser";
	private static final String TEST_PASSWORD = "testpass";

	@Bean
	public ServletRegistrationBean<?> messageDispatcherServlet(ApplicationContext applicationContext) {
		MessageDispatcherServlet servlet = new MessageDispatcherServlet();
		servlet.setApplicationContext(applicationContext);
		servlet.setTransformWsdlLocations(true);
		return new ServletRegistrationBean<>(servlet, "/ws/*");
	}

	@Bean(name = "students")
	public DefaultWsdl11Definition defaultWsdl11Definition(XsdSchema schema) {
		DefaultWsdl11Definition wsdl11Definition = new DefaultWsdl11Definition();
		wsdl11Definition.setPortTypeName("StudentsPort");
		wsdl11Definition.setLocationUri("/ws");
		wsdl11Definition.setTargetNamespace(StudentEndpoint.NAMESPACE_URI);
		wsdl11Definition.setSchema(schema);
		return wsdl11Definition;
	}

	@Bean
	public XsdSchema countriesSchema() {
		return new SimpleXsdSchema(new ClassPathResource(SCHEMA_XSD));
	}

	@Bean
	XwsSecurityInterceptor securityInterceptor() {
		XwsSecurityInterceptor securityInterceptor = new XwsSecurityInterceptor();
		securityInterceptor.setCallbackHandler(callbackHandler());
		securityInterceptor.setSkipValidationIfNoHeaderPresent(true);
		securityInterceptor.setPolicyConfiguration(new ClassPathResource("securityPolicy.xml"));
		return securityInterceptor;
	}

	@Bean
	org.springframework.ws.soap.security.xwss.callback.SimplePasswordValidationCallbackHandler callbackHandler() {
		org.springframework.ws.soap.security.xwss.callback.SimplePasswordValidationCallbackHandler callbackHandler = new org.springframework.ws.soap.security.xwss.callback.SimplePasswordValidationCallbackHandler();
		callbackHandler.setUsersMap(Collections.singletonMap(TEST_USERNAME, TEST_PASSWORD));
		return callbackHandler;
	}

	@Override
	public void addInterceptors(List<EndpointInterceptor> interceptors) {
		interceptors.add(securityInterceptor());
	}
}

@Component
class StudentRepository {

	private static final Map<Long, Student> STUDENTS = new HashMap<>();

	private static final AtomicInteger ID = new AtomicInteger(100);

	public static final List<Long> DEFAULT_STUDENTS = Arrays.asList(100L, 101L, 102L);

	@PostConstruct
	void init() {
		Student student1 = new Student();
		student1.setId(ID.getAndIncrement());
		student1.setFirstName("John");
		student1.setLastName("Doe");
		student1.setGender("Male");
		STUDENTS.put(student1.getId(), student1);

		Student student2 = new Student();
		student2.setId(ID.getAndIncrement());
		student2.setFirstName("Kelly");
		student2.setLastName("Flower");
		student2.setGender("Female");
		STUDENTS.put(student2.getId(), student2);

		Student student3 = new Student();
		student3.setId(ID.getAndIncrement());
		student3.setFirstName("Jason");
		student3.setLastName("Ray");
		student3.setGender("Male");
		STUDENTS.put(student3.getId(), student3);
	}

	public Collection<Student> find(GetStudentsRequest request) {

		Collection<Student> students = STUDENTS.values();

		if (request == null) {
			return students;
		}
		if (request.getId() != null) {
			Student student = STUDENTS.get(request.getId());
			if (student == null) {
				throw new StudentNotFoundException(String.format("student with id '%s' not found", request.getId()));
			}
			return Arrays.asList(student);
		}
		return students.stream()
				.filter(e -> (request.getFirstName() == null || request.getFirstName().isEmpty()
						|| request.getFirstName().equalsIgnoreCase(e.getFirstName()))
						&& (request.getLastName() == null || request.getLastName().isEmpty()
								|| request.getLastName().equalsIgnoreCase(e.getLastName()))
						&& (request.getGender() == null || request.getGender().isEmpty()
								|| request.getGender().equalsIgnoreCase(e.getGender())))

				.collect(Collectors.toList());
	}

	public Student create(CreateStudentRequest request) {
		if (request.getFirstName() == null || "?".equals(request.getFirstName())) {
			throw new DataMissingException("firstName missing in CreateStudentRequest");
		}
		if (request.getLastName() == null || "?".equals(request.getLastName())) {
			throw new DataMissingException("lastName missing in CreateStudentRequest");
		}
		if (request.getGender() == null || "?".equals(request.getGender())) {
			throw new DataMissingException("gender missing in CreateStudentRequest");
		}

		GetStudentsRequest getStudentsRequest = new GetStudentsRequest();
		getStudentsRequest.setFirstName(request.getFirstName());
		getStudentsRequest.setLastName(request.getLastName());
		getStudentsRequest.setGender(request.getGender());

		Collection<Student> find = find(getStudentsRequest);
		if (find.isEmpty()) {
			Student student = new Student();
			student.setId(ID.getAndIncrement());
			student.setFirstName(request.getFirstName());
			student.setLastName(request.getLastName());
			student.setGender(request.getGender());
			STUDENTS.put(student.getId(), student);
			return student;
		} else {
			return find.stream().findFirst().get();
		}
	}

	public Student update(UpdateStudentRequest request) {

		Student student = STUDENTS.get(request.getId());
		if (student == null) {
			throw new StudentNotFoundException(String.format("student with id '%s' not found", request.getId()));
		}

		student.setId(request.getId());
		student.setFirstName(request.getFirstName());
		student.setLastName(request.getLastName());
		student.setGender(request.getGender());

		STUDENTS.put(request.getId(), student);

		return student;
	}

	public boolean delete(DeleteStudentRequest request) {
		if (DEFAULT_STUDENTS.contains(request.getId())) {
			return false;
		}

		Student student = STUDENTS.get(request.getId());
		if (student == null) {
			throw new StudentNotFoundException(String.format("student with id '%s' not found", request.getId()));
		}
		return null != STUDENTS.remove(request.getId());
	}
}

class DataMissingException extends RuntimeException {

	private static final long serialVersionUID = 1L;

	public DataMissingException() {
	}

	public DataMissingException(String msg) {
		super(msg);
	}

}

class StudentNotFoundException extends RuntimeException {

	private static final long serialVersionUID = 1L;

	public StudentNotFoundException() {
	}

	public StudentNotFoundException(String msg) {
		super(msg);
	}

}

@Endpoint
class StudentEndpoint {

	public static final String NAMESPACE_URI = "http://educative.io/soap-automation";

	@Autowired
	private StudentRepository repository;

	@PayloadRoot(namespace = NAMESPACE_URI, localPart = "getStudentsRequest")
	@ResponsePayload
	public GetStudentsResponse getStudents(@RequestPayload GetStudentsRequest request) {
		GetStudentsResponse response = new GetStudentsResponse();
		response.getStudents().addAll(repository.find(request));
		return response;

	}

	@PayloadRoot(namespace = NAMESPACE_URI, localPart = "createStudentRequest")
	@ResponsePayload
	public CreateStudentResponse createStudent(@RequestPayload CreateStudentRequest request) {
		CreateStudentResponse response = new CreateStudentResponse();
		Student student = repository.create(request);
		response.setStudent(student);
		return response;

	}

	@PayloadRoot(namespace = NAMESPACE_URI, localPart = "updateStudentRequest")
	@ResponsePayload
	public UpdateStudentResponse updateStudent(@RequestPayload UpdateStudentRequest request) {

		UpdateStudentResponse response = new UpdateStudentResponse();
		Student student = repository.update(request);
		response.setStudent(student);
		return response;

	}

	@PayloadRoot(namespace = NAMESPACE_URI, localPart = "deleteStudentRequest")
	@ResponsePayload
	public DeleteStudentResponse deleteStudent(@RequestPayload DeleteStudentRequest request) {

		if (StudentRepository.DEFAULT_STUDENTS.contains(request.getId())) {
			throw new RuntimeException("cannot delete default student with id " + request.getId());
		}

		DeleteStudentResponse response = new DeleteStudentResponse();
		boolean isDeleted = repository.delete(request);
		response.setDeleted(isDeleted);
		return response;

	}

}
